# system packages
import sys
import time
import pika
import logging
from scrapy.utils.reqser import request_to_dict
# module packages
from . import connection
from . import picklecompat
import json

logger = logging.getLogger(__name__)


class IQueue(object):
    """Per-spider queue/stack base class"""

    def __init__(self):
        """Init method"""
        raise NotImplementedError

    def __len__(self):
        """Return the length of the queue"""
        raise NotImplementedError

    def push(self, url):
        """Push an url"""
        raise NotImplementedError

    def pop(self, timeout=0):
        """Pop an url"""
        raise NotImplementedError

    def clear(self):
        """Clear queue/stack"""
        raise NotImplementedError


class RabbitMQQueue(IQueue):
    """Per-spider FIFO queue"""

    def __init__(self, connection_url, key, exchange=None, spider=None):
        """Initialize per-spider RabbitMQ queue.

        Parameters:
            connection_url -- rabbitmq connection url
            key -- rabbitmq routing key
        """
        self.key = key
        self.connection_url = connection_url
        self.server = None
        self.serializer = picklecompat
        self.spider = spider
        self.connect()

    def __len__(self):
        """Return the length of the queue"""
        declared = self.channel.queue_declare(self.key, passive=True)
        return declared.method.message_count

    def _try_operation(function):
        """Wrap unary method by reconnect procedure"""

        def wrapper(self, *args, **kwargs):
            try:
                return function(self, *args, **kwargs)
            except Exception as e:
                msg = 'Function %s failed. ErrorMsg... (%s)' %\
                    (str(function), e)
                logger.info(msg)

        return wrapper

    def _encode_request(self, request):
        """Encode a request object"""
        obj = request_to_dict(request, self.spider)
        # return self.serializer.dumps(obj) #指定格式 使用 json @sunny5156
        return json.dumps(obj)

    @_try_operation
    def pop(self, no_ack=False):
        """Pop a message"""
        return self.channel.basic_get(queue=self.key, auto_ack=no_ack)

    @_try_operation
    def ack(self, delivery_tag):
        """Ack a message"""
        self.channel.basic_ack(delivery_tag=delivery_tag)

    @_try_operation
    def push(self, body, headers={}):
        """Push a message"""
        properties = pika.BasicProperties()
        properties.priority = body.priority

        # 处理延时消息
        if '_delay_time' in body.meta:
            headers['x-delay'] = body.meta.get('_delay_time')

        if hasattr(self.spider, 'is_delay_queue') and self.spider.is_delay_queue is True:
            exchange = '{}-delay'.format(self.key)
        else:
            exchange = ''
        properties.headers = headers

        rabbitmq_data = json.dumps({'url':body.url,'params':body.meta}) # 指定格式 使用 json @sunny5156 

        self.channel.basic_publish(exchange=exchange,
                                   routing_key=self.key,
                                   #body=self._encode_request(json_data),
                                   body=rabbitmq_data,
                                   properties=properties)

    def connect(self):
        """Make a connection"""
        if self.server:
            try:
                self.server.close()
            except:
                logger.error('server close error!!!')
                # pass

        self.server = connection.connect(self.connection_url)

        is_delay = False
        if hasattr(self.spider, 'is_delay_queue'):
            is_delay = self.spider.is_delay_queue

        self.channel = connection.get_channel(
            self.server,
            self.key,
            durable=self.spider.settings.get('RABBITMQ_DURABLE', True),
            confirm_delivery=self.spider.settings.get('RABBITMQ_CONFIRM_DELIVERY', True),
            is_delay=is_delay,
            is_lazy=self.spider.settings.get('RABBITMQ_IS_LAZY', True))

    def close(self):
        """Close channel"""
        logger.error('channel is closed!!!!!!!!!!!')
        self.channel.close()

    def clear(self):
        """Clear queue/stack"""
        self.channel.queue_purge(self.key)


__all__ = ['SpiderQueue']



# -*- coding: utf-8 -*-
import pika


def get_channel(connection, queue_name, durable=True, confirm_delivery=True, is_delay=False, is_lazy=True):
    """ Init method to return a prepared channel for consuming
    """
    channel = connection.channel()
    args = {'x-max-priority': 255}
    if is_lazy:
        args['x-queue-mode'] = 'lazy'
    channel.queue_declare(queue=queue_name, durable=durable, arguments=args)
    if confirm_delivery:
        channel.confirm_delivery()

    if is_delay is True:
        exchange_name = "{}-delay".format(queue_name)
        channel.exchange_declare(exchange_name,exchange_type="x-delayed-message",arguments={"x-delayed-type": "direct"})
        channel.queue_bind(queue=queue_name, exchange=exchange_name, routing_key=queue_name)
    return channel


def connect(connection_url):
    """ Create and return a fresh connection
    """
    return pika.BlockingConnection(pika.URLParameters(connection_url))



import sys
import time
import signal
import logging
import pickle
import json

from scrapy.http import Request
from . import connection
from .queue import RabbitMQQueue

logger = logging.getLogger(__name__)


class IScheduler(object):
    """ Base Scrapy scheduler class. """

    def __init__(self):
        raise NotImplementedError

    def open(self, spider):
        """Start scheduling"""
        raise NotImplementedError

    def close(self, reason):
        """Stop scheduling"""
        raise NotImplementedError

    def enqueue_request(self, request):
        """Add request to queue"""
        raise NotImplementedError

    def next_request(self):
        """Pop a request"""
        raise NotImplementedError

    def has_pending_requests(self):
        """Check if queue is not empty"""
        raise NotImplementedError


class Scheduler(IScheduler):
    # TODO: to be extended in future
    @staticmethod
    def _ensure_settings(settings, key):
        if not settings.get(key):
            msg = 'Please set "{}" at settings.'.format(key)
            raise ValueError(msg)


repo_url = 'https://github.com/bigkennel/scrapy-scheduler-rabbitmq'


class RabbitMQScheduler(Scheduler):
    """ A RabbitMQ Scheduler for Scrapy. """
    queue = None
    stats = None

    def __init__(self, connection_url, *args, **kwargs):
        self.connection_url = connection_url
        self.waiting = False
        self.closing = False

    @classmethod
    def from_settings(cls, settings):
        cls._ensure_settings(settings, 'RABBITMQ_CONNECTION_PARAMETERS')
        connection_url = settings.get('RABBITMQ_CONNECTION_PARAMETERS')
        return cls(connection_url)

    @classmethod
    def from_crawler(cls, crawler):
        scheduler = cls.from_settings(crawler.settings)
        scheduler.stats = crawler.stats
        signal.signal(signal.SIGINT, scheduler.on_sigint)
        return scheduler

    def __len__(self):
        return len(self.queue)

    def open(self, spider):
        if not hasattr(spider, '_make_request'):
            msg = 'Method _make_request not found in spider. '
            msg += 'Please add it to spider or see manual at '
            msg += repo_url
            raise NotImplementedError(msg)

        if not hasattr(spider, 'queue_name'):
            msg = 'Please set queue_name parameter to spider. '
            msg += 'Consult manual at ' + repo_url
            raise ValueError(msg)

        self.spider = spider
        self.queue = self._make_queue(spider.queue_name)
        msg_count = len(self.queue)
        if msg_count:
            logger.info(
                'Resuming crawling ({} urls scheduled)'.format(msg_count))
        else:
            logger.info('No items to crawl in {}'.format(self.queue.key))

    def _make_queue(self, key):
        return RabbitMQQueue(self.connection_url, key, spider=self.spider)

    def on_sigint(self, signal, frame):
        self.closing = True

    def close(self, reason):
        try:
            logger.error('scheduler schchannel is closed!!!!!!!!!!!')
            self.queue.close()
        except:
            pass

    def enqueue_request(self, request):
        """ Enqueues request to main queues back
        """
        if self.queue is not None:
            if self.stats:
                self.stats.inc_value('scheduler/enqueued/rabbitmq', spider=self.spider)
            self.queue.push(request)
        return True



    def next_request(self):
        """ Creates and returns a request to fire
        """
        if self.closing:
            return
        no_ack = True if self.spider.settings.get('RABBITMQ_CONFIRM_DELIVERY', True) is False else False

        try:
            mframe, hframe, body = self.queue.pop(no_ack=no_ack)

            if any([mframe, hframe, body]):
                self.waiting = False
                if self.stats:
                    self.stats.inc_value('scheduler/dequeued/rabbitmq',spider=self.spider)

                request = self.spider._make_request(mframe, hframe, body)
                if self.spider.settings.get('RABBITMQ_CONFIRM_DELIVERY', True):
                    request.meta['delivery_tag'] = mframe.delivery_tag

                logger.info('Running request {}'.format(request.url))
                return request
            else:
                if not self.waiting:
                    msg = 'Queue {} is empty. Waiting for messages...'
                    self.waiting = True
                    logger.info(msg.format(self.queue.key))
                return None
        except Exception as e:
            msg = 'Queue {} is error , please reconnect ...'
            logger.info(msg.format(e))
            self.open(self.spider)
            return None

    def has_pending_requests(self):
        return not self.closing


class SaaS(RabbitMQScheduler):
    """ Scheduler as a RabbitMQ service.
    """

    def __init__(self, connection_url, *args, **kwargs):
        super(SaaS, self).__init__(connection_url, *args, **kwargs)

    def ack_message(self, delivery_tag):
        self.queue.ack(delivery_tag)

    def requeue_message(self, body, headers=None):
        self.queue.push(body, headers)



# -*- coding: utf-8-*-
from setuptools import setup, find_packages
setup(
    # 以下为必需参数
    name='scrapy-scheduler-rabbitmq',  # 模块名
    version='1.1.8',  # 当前版本
    description='Rabbitmq for Scrapy Distributed scraping',  # 简短描述
    author='sunny5156',
    author_email='sunny5156@qq.com',
    license='MIT',
    url='https://github.com/bigkennel/scrapy-scheduler-rabbitmq',
    install_requires=[
        'pika',
        'Scrapy>=0.14'
    ],
    packages=['scrapy_scheduler_rabbitmq'],
    package_dir={'': 'src'}
)



# -*- coding: utf-8 -*-
import scrapy
import pickle
import json
from scrapy.utils.reqser import request_from_dict


class RabbitSpider(scrapy.Spider):
    def _make_request(self, mframe, hframe, body):

        data = json.loads(body,encoding="utf-8")
        if self.settings.get("PYPPETEER_ENABLE"):
            data['params']['pyppeteer'] = self.settings.get("PYPPETEER_ENABLE",False)
        request = scrapy.Request(data['url'], callback=self.parse, dont_filter=True,meta=data['params'])

        return request
